# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from gpjax.linops.constant_diagonal_linear_operator import (
    ConstantDiagonalLinearOperator,
)
from gpjax.linops.dense_linear_operator import DenseLinearOperator
from gpjax.linops.diagonal_linear_operator import DiagonalLinearOperator
from gpjax.linops.identity_linear_operator import IdentityLinearOperator
from gpjax.linops.linear_operator import LinearOperator
from gpjax.linops.triangular_linear_operator import (
    LowerTriangularLinearOperator,
    UpperTriangularLinearOperator,
)
from gpjax.linops.utils import (
    identity,
    to_dense,
)
from gpjax.linops.zero_linear_operator import ZeroLinearOperator

__all__ = [
    "LinearOperator",
    "DenseLinearOperator",
    "DiagonalLinearOperator",
    "ConstantDiagonalLinearOperator",
    "IdentityLinearOperator",
    "ZeroLinearOperator",
    "LowerTriangularLinearOperator",
    "UpperTriangularLinearOperator",
    "identity",
    "to_dense",
]
